@@ -6,6 +6,7 @@
 @var XML: to be used to set the return format to XML (SPARQL XML format or RDF/XML, depending on the query type). This is the default.
 @var TURTLE: to be used to set the return format to Turtle
 @var N3: to be used to set the return format to N3 (for most of the SPARQL services this is equivalent to Turtle)
+@var RDF: to be used to set the return RDF Graph
 
 @var POST: to be used to set HTTP POST
 @var GET: to be used to set HTTP GET. This is the default.
@@ -19,36 +20,41 @@
 @authors: U{Ivan Herman<http://www.ivan-herman.net>}, U{Sergio Fernández<http://www.wikier.org>}, U{Carlos Tejo Alonso<http://www.dayures.net>}
 @organization: U{World Wide Web Consortium<http://www.w3.org>} and U{Foundation CTIC<http://www.fundacionctic.org/>}.
 @license: U{W3C® SOFTWARE NOTICE AND LICENSE<href="http://www.w3.org/Consortium/Legal/copyright-software">}
-@requires: U{simplejson<http://cheeseshop.python.org/pypi/simplejson>} package.
-@requires: U{PyXML<http://pyxml.sourceforge.net/>} package.
 @requires: U{RDFLib<http://rdflib.net>} package.
 """
 
 import sys
 import urllib, urllib2
 import re
+import jsonlayer
+import warnings
 from SPARQLWrapper import __agent__
 from SPARQLExceptions import QueryBadFormed, EndPointNotFound
 from SPARQLUtils import deprecated
+from KeyInsensitiveDict import KeyInsensitiveDict
 
 #  Possible output format keys...
 JSON   = "json"
 XML    = "xml"
 TURTLE = "n3"
 N3     = "n3"
-_allowedFormats = [JSON, XML, TURTLE, N3]
+RDF    = "rdf"
+_allowedFormats = [JSON, XML, TURTLE, N3, RDF]
 
 # Possible HTTP methods
 POST = "POST"
 GET  = "GET"
 _allowedRequests = [POST, GET]
 
-# Possible SPARQL query type
+# Possible SPARQL/SPARUL query type
 SELECT     = "SELECT"
 CONSTRUCT  = "CONSTRUCT"
 ASK        = "ASK"
 DESCRIBE   = "DESCRIBE"
-_allowedQueryTypes = [SELECT, CONSTRUCT, ASK, DESCRIBE]
+INSERT     = "INSERT"
+DELETE     = "DELETE"
+MODIFY     = "MODIFY"
+_allowedQueryTypes = [SELECT, CONSTRUCT, ASK, DESCRIBE, INSERT, DELETE, MODIFY]
 
 # Possible output format (mime types) that can be converted by the local script. Unfortunately,
 # it does not work by simply setting the return format, because there is still a certain level of confusion
@@ -60,7 +66,7 @@ _allowedQueryTypes = [SELECT, CONSTRUCT, ASK, DESCRIBE]
 # mime types have just been proposed and not yet widely used...
 _SPARQL_DEFAULT  = ["application/sparql-results+xml", "application/rdf+xml", "*/*"]
 _SPARQL_XML      = ["application/sparql-results+xml"]
-_SPARQL_JSON     = ["application/sparql-results+json", "text/javascript"]
+_SPARQL_JSON     = ["application/sparql-results+json", "text/javascript", "application/json"]
 _RDF_XML         = ["application/rdf+xml"]
 _RDF_N3          = ["text/rdf+n3","application/n-triples","application/turtle","application/n3","text/n3","text/turtle"]
 _ALL             = ["*/*"]
@@ -91,9 +97,9 @@ class SPARQLWrapper :
 	@ivar baseURI: the URI of the SPARQL service
 	"""
 	pattern = re.compile(r"""
-				(?P<base>(\s*BASE\s*[<].*[>])\s*)*
-				(?P<prefixes>(\s*PREFIX\s*.*[:]\s*[<].*[>])\s*)*
-				(?P<queryType>(CONSTRUCT|SELECT|ASK|DESCRIBE))
+                (?P<base>(\s*BASE\s*<.*?>)\s*)*
+                (?P<prefixes>(\s*PREFIX\s+.+:\s*<.*?>)\s*)*
+				(?P<queryType>(CONSTRUCT|SELECT|ASK|DESCRIBE|INSERT|DELETE|MODIFY))
 	""", re.VERBOSE | re.IGNORECASE)
 	def __init__(self,baseURI,returnFormat=XML,defaultGraph=None,agent=__agent__) :
 		"""
@@ -141,7 +147,7 @@ class SPARQLWrapper :
 	def setReturnFormat(self,format) :
 		"""Set the return format. If not an allowed value, the setting is ignored.
 
-		@param format: Possible values: are L{JSON}, L{XML}, L{TURTLE}, L{N3} (constants in this module). All other cases are ignored.
+		@param format: Possible values: are L{JSON}, L{XML}, L{TURTLE}, L{N3}, L{RDF} (constants in this module). All other cases are ignored.
 		@type format: string
 		"""
 		if format in _allowedFormats :
@@ -186,8 +192,8 @@ class SPARQLWrapper :
 			Some SPARQL endpoints allow extra key value pairs.
 			E.g., in virtuoso, one would add C{should-sponge=soft} to the query forcing 
 			virtuoso to retrieve graphs that are not stored in its local database.
-			@param key: key 
-			@type key: string
+			@param name: name 
+			@type name: string
 			@param value: value
 			@type value: string
 			@rtype: bool
@@ -232,6 +238,7 @@ class SPARQLWrapper :
 			return r_queryType
 		else :
 			#raise Exception("Illegal SPARQL Query; must be one of SELECT, ASK, DESCRIBE, or CONSTRUCT")
+			warnings.warn("unknown query type", RuntimeWarning)
 			return SELECT
 
 	def setMethod(self,method) :
@@ -240,6 +247,18 @@ class SPARQLWrapper :
 		"""
 		if method in _allowedRequests : self.method = method
 
+	def setUseKeepAlive(self):
+		"""Make urllib2 use keep-alive.
+		@raise ImportError: when could not be imported urlgrabber.keepalive.HTTPHandler
+		"""
+		try:
+			from urlgrabber.keepalive import HTTPHandler
+			keepalive_handler = HTTPHandler()
+			opener = urllib2.build_opener(keepalive_handler)
+			urllib2.install_opener(opener)
+		except ImportError:
+			warnings.warn("urlgrabber not installed in the system. The execution of this method has no effect.")
+
 	def _getURI(self) :
 		"""Return the URI as sent (or to be sent) to the SPARQL endpoint. The URI is constructed
 		with the base URI given at initialization, plus all the other parameters set.
@@ -255,14 +274,14 @@ class SPARQLWrapper :
 			# However, these processors are (hopefully) oblivious to the parameters they do not understand. 
 			# So: just repeat all possibilities in the final URI. UGLY!!!!!!!
 			for f in _returnFormatSetting: finalQueryParameters[f] = self.returnFormat
-		return self.baseURI + "?" + urllib.urlencode(finalQueryParameters)
+		return self.baseURI + "?" + urllib.urlencode(dict([k, v.encode("utf-8")] for k, v in finalQueryParameters.items()))
 
 	def _createRequest(self) :
 		"""Internal method to create request according a HTTP method. Returns a
 		C{urllib2.Request} object of the urllib2 Python library
 		@return: request
 		"""
-		if self.queryType == SELECT or self.queryType == ASK :
+		if self.queryType == SELECT or self.queryType == ASK or self.queryType == INSERT or self.queryType == DELETE or self.queryType == MODIFY:
 			if self.returnFormat == XML:
 				acceptHeader = ",".join(_SPARQL_XML)
 			elif self.returnFormat == JSON:
@@ -285,9 +304,9 @@ class SPARQLWrapper :
 			# by POST
 			request = urllib2.Request(self.baseURI)
 			request.add_header("Content-Type", "application/x-www-form-urlencoded")
-			request.add_header("'User-Agent", self.agent)
+			request.add_header("User-Agent", self.agent)
 			request.add_header("Accept", acceptHeader)
-			values = { "query" : self.queryString }
+			values = { "query" : self.queryString.encode("utf-8") }
 			data = urllib.urlencode(values)
 			request.add_data(data)
 		else:
@@ -342,7 +361,6 @@ class SPARQLWrapper :
 		res = self.query()
 		return res.convert()
 
-
 #######################################################################################################
 
 class QueryResult :
@@ -384,7 +402,7 @@ class QueryResult :
 		@return: meta information
 		@rtype: dictionary
 		"""
-		return self.response.info()
+		return KeyInsensitiveDict(self.response.info())
 
 	def __iter__(self) :
 		"""Return an iterator object. This method is expected for the inclusion
@@ -396,6 +414,19 @@ class QueryResult :
 		"""Method for the standard iterator."""
 		return self.response.next()
 
+	def setJSONModule(self,module) :
+		"""Set the Python module for encoding JSON data. If not an allowed value, the setting is ignored.
+           JSON modules supported:
+             - ``simplejson``: http://code.google.com/p/simplejson/
+             - ``cjson``: http://pypi.python.org/pypi/python-cjson
+             - ``json``: This is the version of ``simplejson`` that is bundled with the
+               Python standard library since version 2.6
+               (see http://docs.python.org/library/json.html)
+		@param module: Possible values: are L{simplejson}, L{cjson}, L{json}. All other cases raise a ValueError exception.
+		@type module: string
+		"""
+		jsonlayer.use(module)
+
 	def _convertJSON(self) :
 		"""
 		Convert a JSON result into a Python dict. This method can be overwritten in a subclass
@@ -403,15 +434,14 @@ class QueryResult :
 		@return: converted result
 		@rtype: Python dictionary
 		"""
-		#patch to solve bug #2781984
-		jsonlib = None
-		if sys.version_info[0:2] >= (2, 6):
-			import json
-			jsonlib = json
-		else:
-			import simplejson
-			jsonlib = simplejson			
-		return jsonlib.load(self.response)
+		return jsonlayer.decode(self.response.read())
+		# patch to solve bug #2781984
+		# later updated with a trick from http://bob.pythonmac.org/archives/2008/10/02/python-26-released-now-with-json/
+		#try:
+		#	import simplejson as json
+		#except ImportError:
+		#	import json
+		#return json.load(self.response)
 
 	def _convertXML(self) :
 		"""
@@ -420,14 +450,8 @@ class QueryResult :
 		@return: converted result
 		@rtype: PyXlib DOM node
 		"""
-		try :
-			from xml.dom.ext.reader import PyExpat
-			Parser = PyExpat.Reader
-		except ImportError :
-			from xml.dom.ext.reader import Sax2
-			Parser = Sax2.Reader
-		reader = Parser()
-		return reader.fromStream(self.response)
+		from xml.dom.minidom import parse
+		return parse(self.response)
 
 	def _convertRDF(self) :
 		"""
@@ -436,7 +460,10 @@ class QueryResult :
 		@return: converted result
 		@rtype: RDFLib Graph
 		"""
-		from rdflib import ConjunctiveGraph
+		try:
+			from rdflib.graph import ConjunctiveGraph
+		except ImportError:
+			from rdflib import ConjunctiveGraph
 		retval = ConjunctiveGraph()
 		# this is a strange hack. If the publicID is not set, rdflib (or the underlying xml parser) makes a funny
 		#(and, as far as I could see, meaningless) error message...
@@ -450,10 +477,7 @@ class QueryResult :
 		@return: converted result
 		@rtype: string
 		"""
-		retval = ""
-		for l in self.response :
-			retval += l
-		return retval
+		return self.response.read()
 
 	def convert(self) :
 		"""
@@ -465,7 +489,7 @@ class QueryResult :
 
 		@return: the converted query result. See the conversion methods for more details.
 		"""
-		ct = self.response.info()["content-type"]
+		ct = self.info()["content-type"]
 		if True in [ct.find(q) != -1 for q in _SPARQL_XML] :
 			return self._convertXML()
 		elif True in [ct.find(q) != -1 for q in _SPARQL_JSON]  :
@@ -476,5 +500,6 @@ class QueryResult :
 			return self._convertN3()
 		else :
 			# this should cover, well, the rest of the cases...
+			warnings.warn("unknown response content type, returning raw response...", RuntimeWarning)
 			return self.response
 
